part of 'table_widget.dart';

class _AppTableFilterRow extends StatelessWidget {
  final AppButtonConfig? headerAction;
  final String? headerTitle;
  final List<Widget> filters;

  const _AppTableFilterRow({
    this.headerAction,
    this.headerTitle,
    this.filters = const [],
  });

  @override
  Widget build(BuildContext context) {
    return Container(
      padding: const EdgeInsets.symmetric(horizontal: 16),
      constraints: const BoxConstraints(minHeight: 46),
      child: Row(
        crossAxisAlignment: CrossAxisAlignment.start,
        children: <Widget>[
          Expanded(
            child: Wrap(
              spacing: 8,
              runSpacing: 8,
              children: [
                if (filters.isNotEmpty)
                  ...filters
                else if (headerTitle != null)
                  Text(
                    headerTitle!,
                    style: context.textTheme.headlineSmall,
                  ),
              ],
            ),
          ),
          if (headerAction != null) AppButton.fromConfig(headerAction!),
        ].addSpacingBetween(),
      ),
    );
  }
}

sealed class AppFilter<M extends Object> extends StatefulWidget {
  const AppFilter({super.key});

  @override
  State<AppFilter<M>> createState();
}

sealed class _AppFilterState<M extends Object> extends State<AppFilter<M>> {
  @override
  Widget build(BuildContext context) {
    return const Placeholder();
  }
}

class AppTextFilter<M extends Object> extends AppFilter<M> {
  final String label;
  final ValueChanged<String> onChanged;
  final String? initialValue;
  final List<TextInputFormatter>? inputFormatters;

  const AppTextFilter({
    required this.label,
    required this.onChanged,
    this.initialValue,
    this.inputFormatters,
    super.key,
  });

  @override
  State<AppFilter<M>> createState() => _AppTextFilterState<M>();
}

class _AppTextFilterState<M extends Object> extends _AppFilterState<M> {
  @override
  AppTextFilter<M> get widget => super.widget as AppTextFilter<M>;

  late final _textController = TextEditingController(
    text: widget.initialValue ?? '',
  );

  final _overlayController = OverlayPortalController();

  final _link = LayerLink();

  final _focusNode = FocusNode();

  void _focusNodeListener() {
    if (!_focusNode.hasFocus) {
      _setValue();
    }
  }

  void _setValue() {
    widget.onChanged(_textController.text);
    _overlayController.hide();
    context.maybeController<M>()?.reload();
  }

  @override
  void initState() {
    super.initState();

    _focusNode.addListener(_focusNodeListener);
  }

  @override
  void dispose() {
    _focusNode.dispose();
    _textController.dispose();

    super.dispose();
  }

  @override
  Widget build(BuildContext context) {
    return CompositedTransformTarget(
      link: _link,
      child: OverlayPortal(
        controller: _overlayController,
        overlayChildBuilder: (context) {
          return CompositedTransformFollower(
            link: _link,
            showWhenUnlinked: false,
            targetAnchor: Alignment.bottomLeft,
            child: Align(
              alignment: Alignment.topLeft,
              child: SizedBox(
                width: 300,
                child: TapRegion(
                  onTapOutside: (_) => _overlayController.hide(),
                  child: Card(
                    child: Padding(
                      padding: const EdgeInsets.all(8),
                      child: Column(
                        mainAxisSize: MainAxisSize.min,
                        children: [
                          AppTextFormField(
                            controller: _textController,
                            onChanged: widget.onChanged,
                            labelText: widget.label,
                            focusNode: _focusNode,
                            requestFocusOnInitState: true,
                            inputFormatters: widget.inputFormatters,
                          ),
                        ],
                      ),
                    ),
                  ),
                ),
              ),
            ),
          );
        },
        child: ListenableBuilder(
          listenable: _textController,
          builder: (context, child) {
            final hasText = _textController.text.isNotEmpty;

            return FilterChip(
              label: Text(
                [
                  widget.label,
                  if (hasText) _textController.text,
                ].join(': '),
              ),
              selected: hasText,
              showCheckmark: false,
              deleteIconColor: context.colorScheme.onPrimary,
              onSelected: (_) {
                _overlayController.show();
              },
              onDeleted: hasText
                  ? () {
                      _textController.clear();
                      _setValue();
                    }
                  : null,
            );
          },
        ),
      ),
    );
  }
}

class AppBooleanFilter<M extends Object> extends AppFilter<M> {
  final String label;
  final ValueChanged<bool?> onChanged;
  final bool? initialValue;
  final bool triState;

  const AppBooleanFilter({
    required this.label,
    required this.onChanged,
    this.initialValue,
    this.triState = false,
    super.key,
  });

  @override
  State<AppFilter<M>> createState() => _AppBooleanFilterState<M>();
}

class _AppBooleanFilterState<M extends Object> extends _AppFilterState<M> {
  @override
  AppBooleanFilter<M> get widget => super.widget as AppBooleanFilter<M>;

  final _changeDebouncer = Debouncer(duration: const Duration(seconds: 1));

  late bool? _value = widget.initialValue;

  bool? get _nextValue {
    if (_value == null) {
      return true;
    } else if (_value == true && widget.triState) {
      return false;
    } else {
      return null;
    }
  }

  @override
  Widget build(BuildContext context) {
    return FilterChip(
      label: Text(widget.label),
      selected: _value != null,
      showCheckmark: _value == true,
      avatar: _value == false ? const Icon(Icons.remove) : null,
      onSelected: (bool value) {
        setState(() => _value = _nextValue);
        _changeDebouncer.run(() {
          widget.onChanged(_value);
          context.maybeController<M>()?.reload();
        });
      },
    );
  }
}
